/* ----------------------------------------------------------
                          tiny_fuse_reset.c

     YAFR - Yet Another Fuse Reseter

     Programm um mittels "High-Voltage" serial programing
     die Fueses von ATtiny zurueck zu setzen.

     Implementierte ATtiny:

     ATtiny13, ATtiny25/45/85, ATtiny24/44/84

     MCU     : atmega8, atmega48/88/168/328
     F_CPU   : 16000000

     20.12.2019  R. Seelig
   ---------------------------------------------------------- */

#include <util/delay.h>
#include <avr/io.h>

#include "avr_gpio.h"
#include "my_printf.h"
#include "uart_all.h"

#include "fr_hwpins.h"

#define printf           my_printf
#define delay            _delay_ms

// ATtiny Device Signaturen
#define tiny13_id        0x9007
#define tiny24_id        0x910B
#define tiny44_id        0x9207
#define tiny84_id        0x930C
#define tiny25_id        0x9108
#define tiny45_id        0x9206
#define tiny85_id        0x930B


// Defaultwerte Fuses ATtiny13
#define tiny13_lofuse    0x6a
#define tiny13_hifuse    0xff

// ATtiny24/44/84 und ATtiny25/45/85 haben alle dieselben
// Fuses Defaultwerte
#define tiny_lofuse      0x62
#define tiny_hifuse      0xdf
#define tiny_exfuse      0xff

// Kommandowort fuer Fuseauswahl (lo, hi, ext) beim Schreiben
// ist fuer alle unterstuetzten ATtinys gleich
#define lfuse_wrcmd      0x646c
#define hfuse_wrcmd      0x747c
#define efuse_wrcmd      0x666e         // nicht verfuegbar fuer ATtiny13

enum spannungsfolge { pg_0v, pg_5v, pg_12v };

/* -------------------------------------------------
                   my_putchar

     my_putchar is used by my_printf. Place here
     the function, where my_printf should send the
     characters
   ------------------------------------------------- */
void my_putchar(char ch)
{
  uart_putchar(ch);
}

/* -------------------------------------------------------
                       pwm25k_init

     initialisiert auf OC1A (PB1) ein 25 kHz PWM-Signal.
     PB1 muss als Ausgang initialisiert sein
   ------------------------------------------------------- */
void pwm25k_init(void)
{
  TCCR1A= 0;
  TCNT1= 0;

  // Nichtinverter PWM Kanal A ; Modus 10h: Phase korrekt
  TCCR1A= _BV(COM1A1) | _BV(WGM11);

  // Prescaler= 1
  TCCR1B= _BV(WGM13) | _BV(CS10);

  // Werte von 0 bis 320 werden vom PW interpretiert und
  // ergibt somit eine Periodendauer von 40 us
  ICR1= 320;
}

/* -------------------------------------------------------
                       pwm25k_set

      setzt die Pulsdauer der PWM, erlaubte Werte fuer
      value sind 1 .. 319.

      Aufloesung betraegt hierbei 125 ns
   ------------------------------------------------------- */
void pwm25k_set(unsigned int value)
{
  OCR1A= value;
}

/* -------------------------------------------------------
                      connectors_init

     setzt die benoetigten Anschluesse je nach Funktion
     als Aus- oder Eingang.

     Schaltet die gelbe LED ein (als Bereitschafts-LED)
   ------------------------------------------------------- */
void connectors_init(void)
{
  pwmpin_init();  pwr12_init();  pwr5_init();
  sdi_init();     sii_init();    sdo_init();  scl_init();
  ub_init();
  ledgn_init();   ledge_init();  ledrt_init();
  button_init();

  sdi_clr();
  sii_clr();
  scl_clr();

  set_ub_off();

  ledgn_off();
  ledrt_off();
  ledge_off();
}

/* -------------------------------------------------------
                      set_hvoltage

     setzt die Spannung am Target-Controller.

     Werte fuer voltage:

             0 :  0V
             1 :  5V
             2 : 12V
   ------------------------------------------------------- */
void set_hvoltage(uint8_t voltage)
{
  switch (voltage)
  {
    case 0  : set_5v_off(); set_12v_off();  break;
    case 1  : set_5v_on();  set_12v_off();  break;
    case 2  : set_5v_on();  set_12v_on();   break;
    default : break;
  }
}


/* -------------------------------------------------------
                       shiftout11

     schiebt ein Byte mittels 11 !!!! Taktimpulsen aus
     mit MSB bei gleichzeitigem Einlesen des Wertes
     auf SDO.

     Die 11 Taktimpulse sind der Besonderheit des ATtiny
     geschuldet, das vor dem ersten (hoechstwertigsten)
     Bit eine 0 und nach dem letzten Bit zwei weitere
     0 gesendet haben moechte.

     siehe Seite 172 Datasheet ATtiny24/44/84

     sii => Instruction Code
     sdi => Data (u.a. Fuse-Wert)

   ------------------------------------------------------- */
uint8_t shiftout11(uint16_t sdi, uint16_t sii)
{
  uint16_t inbit;
  int16_t z;

  inbit= 0;

  sdi= sdi << 2;
  sii= sii << 2;

  while (!(is_sdo()));                                 // warten bis der Tiny "Bereitschaft" signalisiert
//  printf("\n\r sdi= %d\n\r", sdi);
  for (z= 10; z >= 0; z--)
  {
    if (sdi & (1 << z)) sdi_set(); else sdi_clr();     // serielles Datenbit ausgeben
    if (sii & (1 << z)) sii_set(); else sii_clr();     // serielles Kommandobit ausgeben

    inbit= inbit << 1;
    if (is_sdo()) inbit |= 1;                          // Datenbit vom Tiny lesen
    _delay_us(2);
    scl_set();
    _delay_us(2);
    scl_clr();
  }
  return inbit >> 2;
}


/* -------------------------------------------------------
                      signature_read

     liest die Devicesignatur des Targetcontrollers aus
   ------------------------------------------------------- */
uint16_t signature_read()
{
  uint16_t id = 0;
  uint16_t i;
  uint8_t  val;

  for (i= 1; i< 3; i++)
  {
    shiftout11(0x08, 0x4c);            // SDI = 0x08, SII = 0x4c ==> CMD fuer Signatur lesen
    shiftout11(   i, 0x0c);            // SDI = Bytenr. der Signatur, SII = 0x0c
    shiftout11(0x00, 0x68);            // Bestandteil Signatur lesen
    val = shiftout11(0x00, 0x6c);      // ein Byte der Signatur lesen

    id = (id << 8) + val;              // ein Byte der Signatur in ID einfuegen
  }
  return id;
}


/* -------------------------------------------------------
                           fuse_read

     liest die Fuses aus
   ------------------------------------------------------- */
void fuses_read(uint8_t *lfuse, uint8_t *hfuse, uint8_t *efuse)
{
  shiftout11(0x04, 0x4c);              // fuse lesen
  shiftout11(0x00, 0x68);              // l-fuse selektieren
  *lfuse = shiftout11(0x00, 0x6c);     // l-fuse lesen
  delay(100);

  shiftout11(0x04, 0x4c);              // fuse lesen
  shiftout11(0x00, 0x7a);              // hi-fuse selektieren
  *hfuse = shiftout11(0x00, 0x7e);     // hi-fuse lesen
  delay(100);

  shiftout11(0x04, 0x4c);              // fuse lesen
  shiftout11(0x00, 0x6a);              // e-fuse selektieren
  *efuse = shiftout11(0x00, 0x6e);     // e-fuse lesen
}

/* -------------------------------------------------------
                           fuse_write

     schreibt eine einzelne Fuse

     Bsp.:
       // beschreibe die Lo-Fuse mit dem Wert 0x62
       fuses_write(0x62, lfuse_wrcmd);
   ------------------------------------------------------- */
void fuses_write(uint8_t val, uint16_t fuse_select)
{
  shiftout11(0x40, 0x4c);              // fuse schreiben
  shiftout11(val, 0x2c);               // Wert den die fuse bekommen soll
  shiftout11(0x00, (fuse_select >> 8) & 0xff);
  shiftout11(0x00, (fuse_select & 0xff));
}

/* -------------------------------------------------------
                        fuse_writeall

     schreibt alle Fuses (Lo, Hi, Ext).

     wr_ext:    1 = Extended Fuse wird geschrieben
                    ( fuer alle Tinys ausser ATtiny13)
                0 = Extended Fuse wird nicht geschrieben
                    ( fuer ATtiny13)
   ------------------------------------------------------- */
void fuse_writeall(uint8_t lofuse, uint8_t hifuse, uint8_t extfuse, uint8_t wr_ext)
{
  fuses_write(lofuse, lfuse_wrcmd);
  delay(100);
  fuses_write(hifuse, hfuse_wrcmd);
  delay(100);
  if (wr_ext)
  {
    fuses_write(extfuse, efuse_wrcmd);
    delay(100);
  }
}

/* -------------------------------------------------------
                        uart_gethexzif
      liest ueber die serielle Schnittstelle eine einzelne
      hexadezimale Ziffer ein und interpretiert diese
      als Zahlenwert.

      Bei einer fehlerhaften Eingabe wird der Wert 255
      zurueck geliefert
   ------------------------------------------------------- */
uint8_t uart_gethexzif(void)
{
  uint8_t b;

  b = uart_getchar();

  if ((b>= '0') && (b<= '9')) return (b-'0');
  if ((b>= 'a') && (b<= 'f')) return (b-'a'+10);
  if ((b>= 'A') && (b<= 'F')) return (b-'A'+10);
  return 255;
}

/* -------------------------------------------------------
                        uart_gethex
      liest ueber die serielle Schnittstelle einen
      2-stelligen hexadezimalen Zahlenwert ein
   ------------------------------------------------------- */
uint8_t uart_gethex(void)
{
  uint8_t b, w;

  do
  {
    b= uart_gethexzif();
  }while (b == 255);
  if (b< 10) my_putchar(b+'0'); else my_putchar(b+55);

  w= b << 4;

  do
  {
    b= uart_gethexzif();
  }while (b == 255);
  if (b< 10) my_putchar(b+'0'); else my_putchar(b+55);

  return (w | b);
}


/* ---------------------------------------------------------------------------
                                    M A I N
   --------------------------------------------------------------------------- */
int main(void)
{
  uint16_t id;
  uint8_t  lfuse, hfuse, efuse;
  uint8_t  unknown_id;
  uint8_t  key_pressed, but_pressed;
  uint8_t  blink, ch;

  uart_init();

  pwm25k_init();
  pwm25k_set(40);               // 5 us Puls, 35 us Pause (40 = 1/8 * 320 = 40 us Periodendauer)

  connectors_init();

  while(1)
  {
    set_hvoltage(pg_0v);
    set_ub_off();

    printf("\n\r -----------------------------------\n\r");
    printf("   YAFR - Yet Another Fuse Reseter\n\r");
    printf(" -----------------------------------\n\n\r");
    printf(" Push button to reset device to default values\n\r");
    printf(" or press any key on keyboard to enter manual\n\r");
    printf(" fuse values...\n\r");

    blink= 0;
    do
    {
      // feststellen ob Button oder Taste ueber UART gedrueckt ist
      // und darauf warten, bis eines der beiden Ereignisse eintrifft
      but_pressed= is_button();
      key_pressed= uart_ischar();
      if (blink) { ledge_off(); blink= 0; }
            else { ledge_on(); blink= 1; };
      delay(200);
    } while ( (!(key_pressed)) && (!(but_pressed)) );
    if (key_pressed) ch= uart_getchar();
    while(is_button());
    delay(50);

    // Nacheinander Targetspannung und Programmierspannung einschalten
    delay(600);
    set_ub_on();
    delay(1);
    set_hvoltage(pg_12v);
    delay(100);

    unknown_id= 0;
    id= signature_read();

    printf("\n\r Device - ID : 0x%x \n\r", id);
    printf(" Device      : ");
    switch (id)
    {
      case tiny13_id : { printf("ATtiny13"); break; }
      case tiny25_id : { printf("ATtiny25"); break; }
      case tiny45_id : { printf("ATtiny45"); break; }
      case tiny85_id : { printf("ATtiny85"); break; }
      case tiny24_id : { printf("ATtiny24"); break; }
      case tiny44_id : { printf("ATtiny44"); break; }
      case tiny84_id : { printf("ATtiny84"); break; }
      default: { printf("Unknown device"); unknown_id= 1; break; }
    }
    ledge_off();

    if (!(unknown_id))
    {
      ledgn_on();                 // Anzeige fuer bekanntes Device
      fuses_read(&lfuse, &hfuse, &efuse);
      printf("\n\n\r Fuses       : Lo: 0x%x, Hi: 0x%x, Ext: 0x%x\n\n\r", lfuse, hfuse, efuse);


      if (but_pressed)
      {
        // bei gedruecktem Button in der Hardware, Fuses auf Defaultwerte setzen

        if (but_pressed)
        {
          printf(" reset device to it's default values... \n\r");

          // ATtiny13 hat als einziger Chip andere Defaultvalues
          if (id == tiny13_id) fuse_writeall(0x6a, 0xff, 0xff, 0);
                          else fuse_writeall(0x62, 0xDF, 0xff, 1);
        }
        printf(" ... done, now read fuses for sucess controll ...\n\n\r");
        fuses_read(&lfuse, &hfuse, &efuse);
        printf(" Fuses       : Lo: 0x%x, Hi: 0x%x, Ext: 0x%x\n\n\r", lfuse, hfuse, efuse);
      }
      else
      {
        // war Taste ueber UART gedrueckt, manuelle Eingabe der Fuse-Values ueber
        // UART
        printf("\n\r Enter fuse-values manualy\n\r");
        printf("\n\r Lo-fuse  : "); lfuse= uart_gethex();
        printf("\n\r Hi-fuse  : "); hfuse= uart_gethex();
        if (id != tiny13_id)
        {
          printf("\n\r Ext-Fuse : "); efuse= uart_gethex();
        }
        printf("\n\n\r Hit Enter to flash the fuses or any other key to abort...\n\n\r");
        ch= uart_getchar();
        if (ch= 0x0d)                   // Taste "Enter" gedrueckt ?
        {
          printf(" writing fuse-values to the device...\n\r");

          // ATtiny13 hat als einziger Chip andere Defaultvalues
          if (id == tiny13_id) fuse_writeall(lfuse, hfuse, 0xff, 0);
                          else fuse_writeall(lfuse, hfuse, efuse, 1);
          printf(" ... done, now read fuses for sucess controll ...\n\n\r");
          fuses_read(&lfuse, &hfuse, &efuse);
          printf(" Fuses       : Lo: 0x%x, Hi: 0x%x, Ext: 0x%x\n\n\r", lfuse, hfuse, efuse);

        }
      }
    }
    else
    {
      ledrt_on();
      printf("\n\r");

      // Fehler mit Betaetigen der Taste quittieren
      while(!(is_button()));
      delay(50);
      while(is_button());
      delay(50);
      ledge_on();
      ledrt_off();
    }
    delay(2000);
    ledgn_off();

    delay(100);
    set_hvoltage(pg_0v);
    delay(100);
    set_ub_off();
  }
}
